:: =============================================================================
:: =                Copyright (C) 2015-2018 ProDigital Software.               =
:: =                            All Rights Reserved.                           =
:: =                                                                           =
:: =  This computer program is protected by copyright law and international    =
:: =  treaties.  Unauthorized reproduction or distribution of this program,    =
:: =  or any portion of it, may result in severe civil or criminal penalties,  =
:: =  and will be prosecuted accordingly.                                      =
:: =============================================================================
::
::+=============================================================================
::
:: Name:        DNSListCompiler
:: Ident:       2.19
:: Type:        User Command Task
:: Language:    MS-DOS Batch
:: Index:       Gather lists of addresses and malware domains and compile into a dns_hosts file
::
::
:: Function:
::
::      This script gathers several well-managed lists of malware domains from 
::      sites online and adds it to a static list of host addresses to create a
::      dns_hosts file that will help protect the machines served by the DNS 
::      server.
::
::      With such a list in place your browsing experience will be faster, as
::      ads and trackers will simply not be loaded, and you'll be protected from
::      inadvertently visiting known sites that serve malware.
::      
::      It is possible this will cause some legitimate sites that you visit not 
::      to provide the information you're used to seeing.  Forbes, for example, has
::      added scripting to their web pages that put up messages asking you to disable
::      your ad blocker.  My advice:  If you run into such a site after generating
::      a dns_hosts file with this tool, just browse elsewhere.  There's plenty of
::      other information out there!
::
::
:: Synopsis:
::
::      DNSListCompiler  NoPrompt  TrimWild
::
::      NoPrompt -  (Optional) If specified, will avert the request for confirmation.
::                  This is of course useful for unattended / scheduled operation.
::
::      TrimWild -  (Optional) If specified, will trim dns_hosts list per entries in wild_hosts.
::                  This is optional because it can take tens of minutes of computer time to do.
::
::
::      It is suggested you schedule this job with a command line similar to:
::
::      DNSListCompiler NoPrompt TrimWild >>C:\Some\Path\DNSListCompiler.log 2>&1
::
::
:: Additional Information:
::
::      This script is distributed with several additional tools that are needed for
::      its operation.  Unzip the whole set of tools provided together into one folder 
::      and it should work for you.
::      
::      Malware domain source information includes:
::      
::      http://winhelp2002.mvps.org/hosts.txt
::      http://malware-domains.com/files/domains.zip
::      http://mirror1.malwaredomains.com/files/immortal_domains.txt
::      http://www.malwaredomainlist.com/hostslist/hosts.txt
::      https://adaway.org/hosts.txt
::      http://someonewhocares.org/hosts/hosts
::      http://pgl.yoyo.org/as/serverlist.php?hostformat=hosts&showintro=0&mimetype=plaintext"
::      http://www.quero.at/download/adblock-hosts.zip
::      http://sysctl.org/cameleon/hosts.win
::      
::      The information from above is reduced for uniqueness and appended to your
::      original dns_hosts file.  This can be done multiple times, as the script removes
::      prior information it has automatically added to dns_hosts.
::      
::      
::      If you have a deny-outgoing-connections-by-default firewall setup, you will
::      need to allow the wget.exe program to access the Internet via http: port 80.
::      
::      
::      2.15  Noel Carboni  7/21/2017
::      
::      Added an option to filter non-wildcard entries per wildcarded blacklist.
::      This can make dns_hosts some 11k entries smaller.
::
::
::      2.16  Noel Carboni  8/15/2017
::
::      Changed the time at which the dns_hosts.last and wild_hosts.last files are 
::      stored so that any edits made by hand during the day are shown in the next 
::      day's diff listing.  Now today's results are stored in .last at the end of 
::      the run, while the pre-run results have been moved to .prior.
::
::
::      2.17  Noel Carboni  8/25/2017
::
::      Added preservation of information after the generated section.
::      Corrected detection and exclusion of wild_hosts entries that are 
::      commented out but still include the *. characters in front.
::      Now allows multiple comment characters (e.g., ### x.y.com=0.0.0.0).
::      Made the whitelist generation based on only the content of the portion
::      of the file generated by this script.
::
::
::      2.18  Noel Carboni  9/22/2017
::
::      Corrected some banner text that said HostsCompiler instead of the name
::      of this script.
::
::
::      2.19  Noel Carboni  11/16/2017
::
::      Changed the filename used for the malware-domains.com list from
::      permmalwaredomain.list to malware-domains.list.
::
::-=============================================================================
::
::  Start of process.
::
    @ECHO OFF
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Set the output folder that is to contain the results
::
    IF EXIST C:\DualServer SET _DLC_OUTPUT_FOLDER=C:\DualServer
::
    IF "%_DLC_OUTPUT_FOLDER%" == "" SET _DLC_OUTPUT_FOLDER=%TEMP%
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Ensure all the associated tools are available via the path
::
    SET _DLC_ORIGINAL_PATH=%Path%
    SET Path=%Path%;%~dp0
::
    which tool.cannot.be.found.on.this.system >nul 2>&1
    IF ERRORLEVEL 2 ECHO No which in (.;%Path%)
    IF ERRORLEVEL 2 GOTO TOOLS_ERROR
::
    FOR %%T IN (diff,grep,gsort,sed,ShowTime,sleep,unzip,wget) DO (
        which %%T 2>&1 1>nul
        IF ERRORLEVEL 1 GOTO TOOLS_ERROR
    )
::
    GOTO GOT_TOOLS
::
:TOOLS_ERROR
::
    ECHO.
    ECHO ERROR:  Not all the tools needed could be found either in the path or in the folder with the script.  Unable to continue.
    GOTO CLEAN_UP
::
:GOT_TOOLS
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Get a unique temporary area subfolder to work with
::
    SET _DLC_ORIGINAL_TEMP=%TEMP%
::
    ShowTime "SET TEMP=%_DLC_ORIGINAL_TEMP%\%%H%%M%%S.%%%%03d" >"%_DLC_ORIGINAL_TEMP%\_DLC_Doit.bat"
    CALL "%_DLC_ORIGINAL_TEMP%\_DLC_Doit.bat"
    DEL "%_DLC_ORIGINAL_TEMP%\_DLC_Doit.bat" >nul 2>&1
::
::  Work in the new TEMP area so we can drop temporary files as needed.
::
    MD "%TEMP%" >nul 2>&1
    PUSHD "%TEMP%"
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Check for incorrect parameters.
::
    IF NOT "%3" == "" GOTO DISPLAY_HELP
::
    IF /I "%1" == ""         IF /I "%2" == ""         GOTO INTERPRET_PARAMETERS
    IF /I "%1" == "NOPROMPT" IF /I "%2" == ""         GOTO INTERPRET_PARAMETERS
    IF /I "%1" == "TRIMWILD" IF /I "%2" == ""         GOTO INTERPRET_PARAMETERS
    IF /I "%1" == "NOPROMPT" IF /I "%2" == "TRIMWILD" GOTO INTERPRET_PARAMETERS
    IF /I "%1" == "TRIMWILD" IF /I "%2" == "NOPROMPT" GOTO INTERPRET_PARAMETERS
::
:DISPLAY_HELP
::
    ECHO.
    ECHO ERROR:  Incorrect parameters specified.
    ECHO.
    ECHO Function:
    ECHO.
    ECHO     This script gathers several well-managed lists of malware domains from 
    ECHO     online sources and reduces the list into a dns_hosts file that will help protect 
    ECHO     all the machines served by your DNS server from visiting known malware sites.
    ECHO.
    ECHO.
    ECHO Synopsis:
    ECHO.
    ECHO     DNSListCompiler  NoPrompt  TrimWild
    ECHO.
    ECHO     NoPrompt -  (Optional) If specified, will avert the request for confirmation.
    ECHO                 This is of course useful for unattended / scheduled operation.
    ECHO     
    ECHO     TrimWild -  (Optional) If specified, will trim dns_hosts list per entries in wild_hosts.
    ECHO                 This is optional because it can take tens of minutes of computer time to do.
    ECHO.
::
    GOTO CLEAN_UP
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
:INTERPRET_PARAMETERS
::
    SET _DLC_NOPROMPT_OPTION=
    IF /I "%1" == "NOPROMPT" SET _DLC_NOPROMPT_OPTION=TRUE
    IF /I "%2" == "NOPROMPT" SET _DLC_NOPROMPT_OPTION=TRUE
::
    SET _DLC_TRIMWILD_OPTION=
    IF /I "%1" == "TRIMWILD" SET _DLC_TRIMWILD_OPTION=TRUE
    IF /I "%2" == "TRIMWILD" SET _DLC_TRIMWILD_OPTION=TRUE
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
:CONFIRM_RUN
::
    IF DEFINED _DLC_NOPROMPT_OPTION GOTO START_RUN
::
    SET /P _DLC_CONTINUE="This script will acquire blacklist data online and compile dns_hosts and wild_hosts files.  Continue (Y/N)? "
    IF /I NOT "%_DLC_CONTINUE%" == "Y" GOTO ABORT_RUN
::
    SET _DLC_INTERACTIVE=YES
    GOTO START_RUN
::
:ABORT_RUN
    ECHO Aborted run.
    ECHO.
    GOTO CLEAN_UP
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
:START_RUN
::
::  Ensure we have a working area for list acquisition.
::
    MD "%PROGRAMDATA%\DNSListCompiler" >nul 2>&1
::
    ECHO.
    ECHO Retrieving malware domain information from various sources...
    ECHO.
::
::  Note that this banner is used by the logic that trims out the prior dnshosts data, so think twice before changing it!
::  These banners should be suitable to match things using a SED regular expression and no escaping.
::
    SET "_DLC_BANNER=# Added by DNSListCompiler.bat on "
    SET "_DLC_END_BANNER=# End of entries Added by DNSListCompiler.bat"
::
    ShowTime "%_DLC_BANNER%%%s from:">Banner.txt
    ECHO.>>Banner.txt
::
::  Keep the last received lists from each of the managed list sites.  If the list can't be downloaded, log an error and just use the last one received from that site.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  All known sources of blacklist data are either available directly in TXT format, or are compressed text files in ZIP format.
::
::  If the HOSTS file is available directly in text format, add the line for that source below in the format:
::
::      CALL :_GET_HOST_LIST List_Name "List_URL"
::
::  The List_Name should not contain any spaces, and will be used to create reference copies of the data 
::   from the blacklisted sites from the current and last version of the HOST file obtained from that site, 
::   such as %PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list.
::  
::  If the HOSTS file is only available in ZIP format, add a third parameter to the Function Call, which is 
::   the name of the file inside the ZIP:
::
::      CALL :_GET_HOST_LIST List_Name "List_URL" File_Name_Inside_Zip
::
::  Add as many valid sources here as you like by adding lines in the same format.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
    CALL :_GET_HOST_LIST    mvps               "http://winhelp2002.mvps.org/hosts.txt"
    CALL :_GET_HOST_LIST    malwaredomain      "http://www.malwaredomainlist.com/hostslist/hosts.txt"
    CALL :_GET_HOST_LIST    mwdimmortal        "http://mirror1.malwaredomains.com/files/immortal_domains.txt"
    CALL :_GET_HOST_LIST    adaway             "https://adaway.org/hosts.txt"
    CALL :_GET_HOST_LIST    someonewhocares    "http://someonewhocares.org/hosts/hosts"
    CALL :_GET_HOST_LIST    peterlowe          "http://pgl.yoyo.org/as/serverlist.php?hostformat=hosts&showintro=0&mimetype=plaintext"
    CALL :_GET_HOST_LIST    quero              "http://www.quero.at/download/adblock-hosts.zip" hosts.
    CALL :_GET_HOST_LIST    cameleon           "http://sysctl.org/cameleon/hosts.win"
::
::  Add new sources here by adding lines like the above
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  This gets special treatment since it's delivered in a different format from the others.
::
    SET _DLC_LIST_URL="http://malware-domains.com/files/domains.zip"
::
    ECHO #   %_DLC_LIST_URL% >>Banner.txt
::
    wget -O malware-domains.zip %_DLC_LIST_URL%
    IF ERRORLEVEL 1 GOTO ERROR_MALWARE_DOMAINS_LIST
::
    unzip malware-domains.zip >nul
    IF ERRORLEVEL 1 GOTO ERROR_MALWARE_DOMAINS_LIST
::
    COPY Domains.txt malware-domains.txt >nul
    ECHO.>>malware-domains.txt
::
    sed -r -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "s/^ //g" -e "s/^[0-9]* //g" -e "s/ .*$//g" -e "s/^notice$//g" -e "s/[ ]+$//g" -e "s/\n\n+/\n/g" -e "/^local$/d" -e "/^localhost$/d" -e "/ broadcasthost$/d" -e "/^localhost\.localdomain$/d" -e "/^::/d" -e "/^[[:space:]]*$/d" malware-domains.txt >malware-domains.tmp
    IF ERRORLEVEL 1 GOTO ERROR_MALWARE_DOMAINS_LIST
::
    gsort -f -u malware-domains.tmp >malware-domains.list
::
::  This next line manages the transition from the use of the misleading filename permmalwaredomain.list.  It can be removed in the future.
    IF EXIST "%PROGRAMDATA%\DNSListCompiler\permmalwaredomain.list" IF EXIST malware-domains.list MOVE /Y "%PROGRAMDATA%\DNSListCompiler\permmalwaredomain.list" "%PROGRAMDATA%\DNSListCompiler\malware-domains.last" >nul
::
    IF EXIST "%PROGRAMDATA%\DNSListCompiler\malware-domains.list" COPY "%PROGRAMDATA%\DNSListCompiler\malware-domains.list" "%PROGRAMDATA%\DNSListCompiler\malware-domains.last" >nul
    COPY malware-domains.list "%PROGRAMDATA%\DNSListCompiler\malware-domains.list" >nul
::
    IF NOT EXIST "%PROGRAMDATA%\DNSListCompiler\malware-domains.last" GOTO DONE_MALWARE_DOMAINS_LIST
::
    ECHO malware-domains.list vs. malware-domains.last
    CALL :_COMPARE_LISTS "%PROGRAMDATA%\DNSListCompiler\malware-domains.list" "%PROGRAMDATA%\DNSListCompiler\malware-domains.last"
::
    GOTO DONE_MALWARE_DOMAINS_LIST
::
:ERROR_MALWARE_DOMAINS_LIST
::
    ECHO.
    ECHO ERROR:  malware-domains.com permanent malware domain list could not be retrieved; using existing malware-domains.list
    IF EXIST "%PROGRAMDATA%\DNSListCompiler\malware-domains.list" COPY "%PROGRAMDATA%\DNSListCompiler\malware-domains.list" . >nul
::
:DONE_MALWARE_DOMAINS_LIST
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
    ECHO.>>Banner.txt
::
    ECHO.>EndBanner.txt
    ECHO %_DLC_END_BANNER%>>EndBanner.txt
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  All the files have been retrieved and formatted into just server names
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  We have a list of server names and domain names.
::
::  Note that we remove comment characters off the front of entries because if they're commented out then chances are
::  they've been whitelisted by the user and we want to exclude them from the added dataset.
::
    ECHO # The following are whitelisted entries found commented-out in the previously generated portion >WhiteListBanner.txt
    ECHO # of this file.  These are excluded from the generated blacklist.>>WhiteListBanner.txt
    ECHO.>>WhiteListBanner.txt
::
::  Isolate just the server names by removing comments, IPv6, blank lines, and any IP address prefixes.
::
    sed -r -n -e "/^%_DLC_BANNER%/q;p" "%_DLC_OUTPUT_FOLDER%\dns_hosts."  >BaseServerFile.txt 2>nul
    sed -r -n -e "/^%_DLC_BANNER%/q;p" "%_DLC_OUTPUT_FOLDER%\wild_hosts." >BaseDomainFile.txt 2>nul
::
    sed -r -e "/^%_DLC_END_BANNER%/,$!d" -e "/^%_DLC_END_BANNER%/d" "%_DLC_OUTPUT_FOLDER%\dns_hosts."  >TrailingServerFile.txt 2>nul
    sed -r -e "/^%_DLC_END_BANNER%/,$!d" -e "/^%_DLC_END_BANNER%/d" "%_DLC_OUTPUT_FOLDER%\wild_hosts." >TrailingDomainFile.txt 2>nul
::
::  Extract just the sections that are generated by this script for use in deteriming the whitelists.  This prevents commented-out entries
::  from being migraded from the base or trailing portions of the file into the whitelist.  Someone else is managing those sections.
::
    sed -r -n -e "/^%_DLC_BANNER%/,$!d" -e "/^%_DLC_BANNER%/d" -e "/^%_DLC_END_BANNER%/q;p" "%_DLC_OUTPUT_FOLDER%\dns_hosts."  >GeneratedServerSection.txt 2>nul
    sed -r -n -e "/^%_DLC_BANNER%/,$!d" -e "/^%_DLC_BANNER%/d" -e "/^%_DLC_END_BANNER%/q;p" "%_DLC_OUTPUT_FOLDER%\wild_hosts." >GeneratedDomainSection.txt 2>nul
::
    grep -E "^#+[[:space:]]*[-a-zA-Z0-9_.]+\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+"  GeneratedServerSection.txt | gsort -f -u >WhiteServerList.txt 2>nul
    grep -E "^#+[[:space:]]*[-a-zA-Z0-9_.*]+\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+" GeneratedDomainSection.txt | gsort -f -u >WhiteDomainList.txt 2>nul
::
::  Don't, however, include whitelisted entries, 
::   OR exact matches from entries in the user's sections of the lists - no sense in duplicating them in the generated lists.
::
    sed -r -e "s/^#+[[:space:]]*([-a-zA-Z0-9_.]+\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)/\1/g" -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "s/ +$//" -e "/^::/d" -e "/^ *$/d" -e "s/\n\n+/\n/g" -e "s/\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+//g" WhiteServerList.txt >ServerExclusions.tmp
    sed -r -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "/^::/d" -e "/^ *$/d" -e "s/\n\n+/\n/g" -e "s/ +$//" -e "s/\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+//g" BaseServerFile.txt >>ServerExclusions.tmp
    sed -r -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "/^::/d" -e "/^ *$/d" -e "s/\n\n+/\n/g" -e "s/ +$//" -e "s/\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+//g" TrailingServerFile.txt >>ServerExclusions.tmp
::
    sed -r -e "s/^#+[[:space:]]*([-a-zA-Z0-9_.*]+\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)/\1/g" -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "s/ +$//" -e "/^::/d" -e "/^ *$/d" -e "s/\n\n+/\n/g" -e "s/\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+//g" -e "s/^\*\.//g" WhiteDomainList.txt >DomainExclusions.tmp
    sed -r -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "/^::/d" -e "/^ *$/d" -e "s/\n\n+/\n/g" -e "s/ +$//" -e "s/\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+//g" -e "s/^\*\.//g" BaseDomainFile.txt >>DomainExclusions.tmp
    sed -r -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "/^::/d" -e "/^ *$/d" -e "s/\n\n+/\n/g" -e "s/ +$//" -e "s/\=[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+//g" -e "s/^\*\.//g" TrailingDomainFile.txt >>DomainExclusions.tmp
::
::  Create a sorted exclusion list that can be compared to the new entries being added.
::
    gsort -f -u ServerExclusions.tmp >ServerExclusions.srt
    gsort -f -u DomainExclusions.tmp >DomainExclusions.srt
::
::  Get all just-generated lists into one file.
::
    COPY /B *.list NewServerData.tmp >nul
::
::  Sort and eliminate duplicates in the list of hosts downloaded.
::
    gsort -f -u NewServerData.tmp >NewServerData.srt
    IF ERRORLEVEL 1 GOTO EDIT_FAIL
::
::  Make sure we're not adding any entries that are already in the user's sections or whitelist.
::
    diff -i -u ServerExclusions.srt NewServerData.srt | grep -E "^\+" | sed -r -e "1d" -e "s/^.//" >NewServerData.list
::
::  Split out a second list of just domain names and trim it of duplicates and whitelisted entries as well.
::
    grep -E "^[^.]+\.[^.]+$" NewServerData.list >NewDomainData.srt
    diff -i -u DomainExclusions.srt NewDomainData.srt | grep -E "^\+" | sed -r -e "1d" -e "s/^.//" >NewDomainData.list
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Save a copy of the dns_hosts and wild_hosts files we're starting with for possible restoral on error
::
    IF EXIST "%_DLC_OUTPUT_FOLDER%\dns_hosts." COPY "%_DLC_OUTPUT_FOLDER%\dns_hosts." "%_DLC_OUTPUT_FOLDER%\dns_hosts.prior" >nul 2>&1
    IF ERRORLEVEL 1 GOTO EDIT_FAIL
::
    IF EXIST "%_DLC_OUTPUT_FOLDER%\wild_hosts." COPY "%_DLC_OUTPUT_FOLDER%\wild_hosts." "%_DLC_OUTPUT_FOLDER%\wild_hosts.prior" >nul 2>&1
    IF ERRORLEVEL 1 GOTO RESTORE_DNS_HOSTS_ONLY
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
    ECHO.>BlackListBanner.txt
    ECHO # The generated blacklist follows.   Enjoy your enhanced security!>>BlackListBanner.txt
    ECHO.>>BlackListBanner.txt
    ECHO #  Note:  If you want to exclude any entry from the blacklist, comment it out by placing>>BlackListBanner.txt
    ECHO #         a single # character in front of it.  The name resolution for that site will>>BlackListBanner.txt
    ECHO #         then be allowed and it will be excluded from future DNSListCompiler runs.>>BlackListBanner.txt
    ECHO.>>BlackListBanner.txt
::
::  At this point we have a clean set of unique, sorted server names.  Set dns_hosts file formatting.
::
    sed -r -e "s/$/=0.0.0.0/" NewServerData.list >NewServerData.txt
::
::  In the dns_hosts file determine if anything was previously added by this script.
::  If we can't find anything previously added by this script, don't do a comparison for the log.
::
    grep "^%_DLC_BANNER%" "%_DLC_OUTPUT_FOLDER%\dns_hosts." >nul 2>&1
    IF NOT ERRORLEVEL 1 SET _DLC_DO_DNS_DIFF=YES
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Set wild_hosts file formatting, prefixing the domain names with asterisk dot.
::
    sed -r -e "s/^/*./" -e "s/$/=0.0.0.0/" NewDomainData.list >NewDomainData.txt
::
::  In the wild_hosts file determine if anything was previously added by this script.
::  If we can't find anything previously added by this script, don't do a comparison for the log.
::
    grep "^%_DLC_BANNER%" "%_DLC_OUTPUT_FOLDER%\wild_hosts." >nul 2>&1
    IF NOT ERRORLEVEL 1 SET _DLC_DO_WILD_DIFF=YES
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  TODO:  Come up with a way of preserving everything after the end banner.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Create the final wild_hosts file - /B switch averts a hex 1A from being tacked on the end
::
    COPY /B BaseDomainFile.txt+Banner.txt+WhiteListBanner.txt+WhiteDomainList.txt+BlackListBanner.txt+NewDomainData.txt+EndBanner.txt+TrailingDomainFile.txt wild_hosts. >nul
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
    IF NOT DEFINED _DLC_TRIMWILD_OPTION GOTO CREATE_DNS_HOSTS
::
::  If the TrimWild option has been specified, trim NewServerData.txt entries that are already covered by the wild_hosts file.
::   This reduces the list size markedly and should improve efficiency, but can take tens of minutes to run.
::
    sed -r -e "s/#.*$//g" -e "/^[[:space:]]*$/d" -e "s/\=.*$//g" -e "s/\./\\\./" -e "s/\*/\.\*/g" -e "s/^/\/\^/g" -e "s/$/\=\/d/g" wild_hosts. >delete_wild_entries.sed
    sed -f delete_wild_entries.sed NewServerData.txt >NewServerData.trimmed
    COPY NewServerData.trimmed NewServerData.txt
::
:CREATE_DNS_HOSTS
::
::  Create the final dns_hosts file - /B switch averts a hex 1A from being tacked on the end
::
    COPY /B BaseServerFile.txt+Banner.txt+WhiteListBanner.txt+WhiteServerList.txt+BlackListBanner.txt+NewServerData.txt+EndBanner.txt+TrailingServerFile.txt dns_hosts. >nul
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Now copy the finished files into position.  Try several times to work around any momentary file conflicts.
::
    COPY dns_hosts. "%_DLC_OUTPUT_FOLDER%\dns_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_HOSTS_CREATED
::
    sleep 1
    COPY dns_hosts. "%_DLC_OUTPUT_FOLDER%\dns_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_HOSTS_CREATED
::
    sleep 10
    COPY dns_hosts. "%_DLC_OUTPUT_FOLDER%\dns_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_HOSTS_CREATED
::
    sleep 60
    COPY dns_hosts. "%_DLC_OUTPUT_FOLDER%\dns_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_HOSTS_CREATED
::
    GOTO RESTORE_HOSTS
::
:NEW_HOSTS_CREATED
::
    COPY wild_hosts. "%_DLC_OUTPUT_FOLDER%\wild_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_WILDHOSTS_CREATED
::
    sleep 1
    COPY wild_hosts. "%_DLC_OUTPUT_FOLDER%\wild_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_WILDHOSTS_CREATED
::
    sleep 10
    COPY wild_hosts. "%_DLC_OUTPUT_FOLDER%\wild_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_WILDHOSTS_CREATED
::
    sleep 60
    COPY wild_hosts. "%_DLC_OUTPUT_FOLDER%\wild_hosts." >nul
    IF NOT ERRORLEVEL 1 GOTO NEW_WILDHOSTS_CREATED
::
    GOTO RESTORE_HOSTS
::
:NEW_WILDHOSTS_CREATED
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Success!
::
::  Summarize what's changed in the dns_hosts file.
::
    ECHO.
::
    IF EXIST "%_DLC_OUTPUT_FOLDER%\dns_hosts.prior" IF "%_DLC_DO_DNS_DIFF%" == "YES" GOTO DNS_DIFF
::
::  First time run - just summarize
::
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
    grep -c $ NewServerData.list | sed -r -e "s/$/ new entries added to your dns_hosts file.  Enjoy your improved browsing and security!/"
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
    GOTO DNS_DIFF_DONE
::
:DNS_DIFF
::
::  Not first time run - give details about what's been changed
::
    ECHO dns_hosts vs. dns_hosts.last
    CALL :_COMPARE_LISTS "%_DLC_OUTPUT_FOLDER%\dns_hosts." "%_DLC_OUTPUT_FOLDER%\dns_hosts.last"
::
:DNS_DIFF_DONE
::
    grep -c $ dns_hosts | sed -r -e "s/^/dns_hosts total entries:  /"
    ECHO.
    DIR %_DLC_OUTPUT_FOLDER%\dns_hosts.*
    ECHO.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Now summarize what's changed in the wild_hosts file.
::
    ECHO.
::
    IF EXIST "%_DLC_OUTPUT_FOLDER%\wild_hosts.last" IF "%_DLC_DO_WILD_DIFF%" == "YES" GOTO WILD_DIFF
::
::  First time run - just summarize
::
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
    grep -c $ NewDomainData.list | sed -r -e "s/$/ new entries added to your wild_hosts file.  Enjoy your improved browsing and security!/"
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
    GOTO WILD_DIFF_DONE
::
:WILD_DIFF
::
::  Not first time run - give details about what's been changed
::
    ECHO wild_hosts vs. wild_hosts.last
    CALL :_COMPARE_LISTS "%_DLC_OUTPUT_FOLDER%\wild_hosts." "%_DLC_OUTPUT_FOLDER%\wild_hosts.last"
::
:WILD_DIFF_DONE
::
    grep -c $ wild_hosts | sed -r -e "s/^/wild_hosts total entries:  /"
    ECHO.
    DIR %_DLC_OUTPUT_FOLDER%\wild_hosts.*
    ECHO.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Save a copy of the auto-generated dns_hosts and wild_hosts files we're starting with for comparison next run.
::   Doing this now vs. at the start of the next run ensures that the next day's diff shows not only automated 
::   changes but any changes done by hand as well.
::
    COPY "%_DLC_OUTPUT_FOLDER%\dns_hosts." "%_DLC_OUTPUT_FOLDER%\dns_hosts.last" >nul 2>&1
    COPY "%_DLC_OUTPUT_FOLDER%\wild_hosts." "%_DLC_OUTPUT_FOLDER%\wild_hosts.last" >nul 2>&1
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  Now restart the server to load the new file.
::
    TaskList /SVC /FO:CSV /NH >TaskList.txt 2>&1
    grep "DualServer.exe" TaskList.txt >nul 2>&1
    IF ERRORLEVEL 1 GOTO LOG_FINAL
::
    ECHO.
    NET stop DUALServer
    NET start DUALServer
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
    GOTO LOG_FINAL
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
::  If an error occurred in creating the actual dns_hosts and wild_hosts files, restore the ones we started with
::
:RESTORE_HOSTS
::
    COPY "%_DLC_OUTPUT_FOLDER%\wild_hosts.prior" "%_DLC_OUTPUT_FOLDER%\wild_hosts."
::
:RESTORE_DNS_HOSTS_ONLY
::
    COPY "%_DLC_OUTPUT_FOLDER%\dns_hosts.prior" "%_DLC_OUTPUT_FOLDER%\dns_hosts."
::
:EDIT_FAIL
::
    ECHO.
    ECHO ERROR:  Could not process one or more of the data files.  This is as far as we got:
    DIR
    ECHO.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
:LOG_FINAL
::
    ECHO.
    ShowTime DNSListCompiler.bat finished on %%s.
    ECHO.
    ECHO ============================================================================================================================================
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
:CLEAN_UP
::
    IF NOT "%_DLC_ORIGINAL_PATH%" == "" SET Path=%_DLC_ORIGINAL_PATH%
::
    IF "%_DLC_ORIGINAL_TEMP%" == "" GOTO TEMP_CLEANUP_DONE
    POPD
    RD /S /Q %TEMP% >nul
    SET TEMP=%_DLC_ORIGINAL_TEMP%
::
:TEMP_CLEANUP_DONE
::
    IF "%_DLC_INTERACTIVE%" == "YES" PAUSE
::
::  Delete environment variables, all of which start with "_DLC_"
::
    FOR /F "tokens=1* delims==" %%G IN ('"SET "_DLC_" 2>nul"') DO (SET "%%G=")
::
    GOTO END_OF_PROCEDURE
::
::==============================================================================
::
::  Callable function to retrieve hosts data from a site online and format it.
::
::  Keep the last received lists from each of the managed list sites.  
::  If the list can't be downloaded, log an error and just use the last one received from that site.
::
:: = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
::
:_GET_HOST_LIST _DLC_LIST_NAME _DLC_LIST_URL [_DLC_FILE_NAME_INSIDE_ZIP]
::
    SET _DLC_LIST_NAME=%1
    SET _DLC_LIST_URL=%2
::
    ECHO #   %_DLC_LIST_URL%>>Banner.txt
::
    IF '%3'=='' (SET _DLC_EXT=txt) ELSE ((SET _DLC_EXT=zip) & (SET _DLC_FILE_NAME_INSIDE_ZIP=%3))
::
    wget --no-check-certificate -O %_DLC_LIST_NAME%.%_DLC_EXT% %_DLC_LIST_URL%
    IF ERRORLEVEL 1 GOTO ERROR_GET_HOST_LIST
    IF "%_DLC_EXT%" == "txt" GOTO CONTINUE_GET_HOST_LIST
::
    unzip %_DLC_LIST_NAME%.%_DLC_EXT% >nul
    IF ERRORLEVEL 1 GOTO ERROR_GET_HOST_LIST
::
    COPY %_DLC_FILE_NAME_INSIDE_ZIP% %_DLC_LIST_NAME%.txt >nul
::
:CONTINUE_GET_HOST_LIST
::
    ECHO.>>%_DLC_LIST_NAME%.txt
::
::  Isolate just IPv4 server names.  Strip off any addresses from the beginning and port numbers or bracketed comments from the end.
::
    sed -r -e "s/#.*$//g" -e "s/[[:space:]]+/ /g" -e "s/^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+ //g" -e "s/:.*$//g" -e "s/\[.*$//g" -e "s/[ ]+$//g" -e "s/\n\n+/\n/g" -e "/^local$/d" -e "/^localhost$/d" -e "/^broadcasthost$/d" -e "/^localhost\.localdomain$/d" -e "/^[[:space:]]*$/d" %_DLC_LIST_NAME%.txt >%_DLC_LIST_NAME%.tmp
    IF ERRORLEVEL 1 GOTO ERROR_GET_HOST_LIST
::
    gsort -f -u %_DLC_LIST_NAME%.tmp >%_DLC_LIST_NAME%.list
::
    IF EXIST "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list" COPY "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list" "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.last" >nul
    COPY %_DLC_LIST_NAME%.list "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list" >nul
::
    IF NOT EXIST "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.last" GOTO DONE_GET_HOST_LIST
::
    ECHO %_DLC_LIST_NAME%.list vs. %_DLC_LIST_NAME%.last
    CALL :_COMPARE_LISTS "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list" "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.last"
::
    GOTO DONE_GET_HOST_LIST
::
:ERROR_GET_HOST_LIST
::
    ECHO.
    ECHO ERROR:  %_DLC_LIST_NAME% hosts list could not be retrieved; using existing %_DLC_LIST_NAME%.list
    IF EXIST "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list" COPY "%PROGRAMDATA%\DNSListCompiler\%_DLC_LIST_NAME%.list" .
::
:DONE_GET_HOST_LIST
::
    EXIT /B
::
::==============================================================================
::
::  Callable function to compare lists and log details and summary counts.
::
::  Remember to include quotes around the arguments.
::
:_COMPARE_LISTS _DLC_NEW_LIST _DLC_OLD_LIST
::
    SET _DLC_NEW_LIST=%1
    SET _DLC_OLD_LIST=%2
::
    IF NOT EXIST %_DLC_OLD_LIST% COPY nul %_DLC_OLD_LIST%
::
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
    diff -y --suppress-common-lines --width=140 %_DLC_NEW_LIST% %_DLC_OLD_LIST% >list.dif
::
::  Note that this counts lines with | that are not comments as both an add and a remove
::
    FOR /F %%C IN ('grep -Ec "^[[:space:]]*[^#[:space:]].*[<|]" list.dif') DO (SET "_DLC_ADDED=%%C")
    FOR /F %%C IN ('grep -Ec "[>|][[:space:]]*[^#[:space:]]" list.dif') DO (SET "_DLC_REMOVED=%%C")
    IF "%_DLC_ADDED%" == "0" IF "%_DLC_REMOVED%" == "0" GOTO COMPARE_SUMMARY
::
    TYPE list.dif
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
::
:COMPARE_SUMMARY
::
    ECHO                            Added:  %_DLC_ADDED%                                                           Removed:  %_DLC_REMOVED%
::
    ECHO --------------------------------------------------------------------------------------------------------------------------------------------
    ECHO.
::
    EXIT /B
::
::==============================================================================
::
:END_OF_PROCEDURE
::
::==============================================================================
::
::  End of file.
::
::==============================================================================
